/******************************************************************************
 * Export.cpp
 ******************************************************************************
 $Header:: /N-Patch/Max3.1Exporter/Export.cpp 4     4/13/01 4:31p Avlachos    $
 ******************************************************************************
 * Created by Alex Vlachos (AVlachos@ati.com) on 4-12-2001
 * (C) 2000 ATI Research, Inc.  All rights reserved.
 ******************************************************************************/

// INCLUDES ===================================================================
#include "NPatchExp.h"
#include "AtiMAXVertexNormals.h"

// GLOBALS ====================================================================
static float FindVertexAngle(Mesh *mesh, int face, int corner);
static char8 gIniFileName[64] = "";
static int32 gFirstTime = TRUE;

//=============================================================================
// Dummy function for progress bar
//=============================================================================
DWORD WINAPI fn (LPVOID arg)
{
   return (0);
}

//=============================================================================
void SuErrorLog (const char8 *szFmt, ...)
{
   char8 sz[4096], szMsg[4096];

   va_list va;
   va_start(va, szFmt);
   vsprintf(szMsg, szFmt, va);
   va_end(va);

   sprintf(sz, "%s ", szMsg);

   sz[sizeof(sz)-1] = '\0';

#ifdef _DEBUG
   OutputDebugString(sz);
#endif

   MessageBox (NULL, sz, "Error!", MB_OK);
}

//=============================================================================
void SuIniLog (const char8 *szFmt, ...)
{
   FILE *fp;
   char8 sz[4096], szMsg[4096];

   va_list va;
   va_start(va, szFmt);
   vsprintf(szMsg, szFmt, va);
   va_end(va);

   sprintf(sz, "%s", szMsg);

   sz[sizeof(sz)-1] = '\0';

#ifdef _DEBUG
   OutputDebugString(sz);
#endif

   if (gFirstTime == TRUE)
   {
      fp = fopen(gIniFileName, "w");
      if (fp == NULL)
         return;
      gFirstTime = FALSE;
   }
   else
   {
      fp = fopen(gIniFileName, "a");
      if (fp == NULL)
         return;
   }

   fprintf (fp, sz);
   fclose (fp);
   fflush (stdout);
}

//=============================================================================
void SuIniLog (char8 *szFmt)
{
   SuIniLog((const char8 *)szFmt);
}

//=============================================================================
// Start the exporter!
// This is the real entrypoint to the exporter. After the user has selected
// the filename (and he's prompted for overwrite etc.) this method is called.
//=============================================================================
int SushiOut::DoExport(const TCHAR *aFileName, ExpInterface *ei, Interface *interfacePointer, BOOL suppressPrompts, DWORD options)
{
   int i;
   char fileName[128];

   ip = interfacePointer; //Grab the interface pointer
   gFirstTime = TRUE;

   /****************/
   /* Make x.M x.m */
   /****************/
   strcpy (fileName, aFileName);
   if (strlen(fileName) >= 5)
   {
      fileName[strlen(fileName)-1] = 'm';
   }

   /**************************/
   /* Start the progress bar */
   /**************************/
   ip->ProgressStart(GetString(IDS_PROGRESS_MSG), TRUE, fn, NULL);
   ip->ProgressUpdate(0); 
   mCurNode = 0;
   mTotalNodeCount = 0;

   /*************************************/
   /* Count all entities being exported */
   /*************************************/
   mNumPolygons = 0;
   mNumLights = 0;
   mPolygonArray = NULL;
   for (int idx=0; idx<ip->GetRootNode()->NumberOfChildren(); idx++)
   {
      CountAllByNodeEnum(ip->GetRootNode()->GetChildNode(idx));

      if (ip->GetCancel())
         return 0;
   }

   /*******************/
   /* Allocate memory */
   /*******************/
   if (mNumPolygons > 0)
   {
      mPolygonArray = (NPatchPolygon *) malloc (sizeof(NPatchPolygon)*mNumPolygons);
      if (mPolygonArray == NULL)
      {
         SuErrorLog("Error allocatin %d bytes of memory for mPolygonArray!", (sizeof(NPatchPolygon)*mNumPolygons));
         ip->SetCancel(TRUE);
      }
      mNumPolygons = 0;
   }

   /**************************************************/
   /* Export everything else into AtiScene structure */
   /**************************************************/
   for (i=0; i<ip->GetRootNode()->NumberOfChildren(); i++)
   {
      if (ip->GetCancel())
         break;
      NodeEnum(ip->GetRootNode()->GetChildNode(i));
   }

   /**********************************/
   /* Create vertex and index arrays */
   /**********************************/
   if (!ip->GetCancel())
   {
      mNumVertices = 0;
      mVertexArray = NULL;

      mNumIndices = 0;
      mIndexArray = NULL;

      for (i=0; i<mNumPolygons; i++) //For each polygon
      {
         int32 j;
         int32 k;
         NPatchPolygon *poly = &(mPolygonArray[i]);

         /********************/
         /* Add this polygon */
         /********************/
         AddVertexToArray(poly, 0);
         AddVertexToArray(poly, 1);
         AddVertexToArray(poly, 2);

         /***************************************************/
         /* Find crease edges and write out crease polygons */
         /***************************************************/
         for (j=i+1; j<mNumPolygons; j++) //For each polygon ahead of this in the list
         {
            int32 currentEdge;
            int32 poly1v1 = -1;
            int32 poly1v2 = -1;
            int32 poly2v1 = -1;
            int32 poly2v2 = -1;
            NPatchPolygon *poly1 = &(mPolygonArray[i]);
            NPatchPolygon *poly2 = &(mPolygonArray[j]);

            for (currentEdge=0; currentEdge<3; currentEdge++) //For each edge of this polygon
            {
               if (currentEdge == 0)
               {
                  poly1v1 = 0;
                  poly1v2 = 1;
               }
               else if (currentEdge == 1)
               {
                  poly1v1 = 1;
                  poly1v2 = 2;
               }
               else if (currentEdge == 2)
               {
                  poly1v1 = 2;
                  poly1v2 = 0;
               }

               /**************************/
               /* Find matching vertices */
               /**************************/
               poly2v1 = -1;
               poly2v2 = -1;
               for (k=0; k<3; k++) //For each vertex in poly2
               {
                  if ((poly1->vertex[poly1v1].x == poly2->vertex[k].x) &&
                      (poly1->vertex[poly1v1].y == poly2->vertex[k].y) &&
                      (poly1->vertex[poly1v1].z == poly2->vertex[k].z))
                  {
                     poly2v1 = k;
                     break;
                  }
               }

               for (k=0; k<3; k++) //For each vertex in poly2
               {
                  if ((poly1->vertex[poly1v2].x == poly2->vertex[k].x) &&
                      (poly1->vertex[poly1v2].y == poly2->vertex[k].y) &&
                      (poly1->vertex[poly1v2].z == poly2->vertex[k].z))
                  {
                     poly2v2 = k;
                     break;
                  }
               }

               /*******************************/
               /* Check for different normals */
               /*******************************/
               if ((poly2v1 != -1) && (poly2v2 != -1)) //If we found a shared edge
               {
                  /***************************************/
                  /* Create crease polygons where needed */
                  /***************************************/
                  if ((poly1->normal[poly1v1].x != poly2->normal[poly2v1].x) ||
                      (poly1->normal[poly1v1].y != poly2->normal[poly2v1].y) ||
                      (poly1->normal[poly1v1].z != poly2->normal[poly2v1].z))
                  {
                     //Polygon1: poly1v2, poly1v1, poly2v1
                     AddVertexToArray(poly1, poly1v2);
                     AddVertexToArray(poly1, poly1v1);
                     AddVertexToArray(poly2, poly2v1);
                  }

                  if ((poly1->normal[poly1v2].x != poly2->normal[poly2v2].x) ||
                      (poly1->normal[poly1v2].y != poly2->normal[poly2v2].y) ||
                      (poly1->normal[poly1v2].z != poly2->normal[poly2v2].z))
                  {
                     //Polygon2: poly1v2, poly2v1, poly2v2
                     AddVertexToArray(poly1, poly1v2);
                     AddVertexToArray(poly2, poly2v1);
                     AddVertexToArray(poly2, poly2v2);
                  }

                  break;
               }
            }
         }
      }

      /**********************/
      /* Write file to disk */
      /**********************/
      FILE *fp;
      fp = fopen(fileName, "w");
      if (fp == NULL)
      {
         SuErrorLog("Error opening file %s!", fileName);
      }
      else
      {
         //Write vertices
         for (i=0; i<mNumVertices; i++) //For each vertex
         {
            NPatchVertex *vert = &(mVertexArray[i]);
            fprintf(fp, "V %f %f %f %f %f %f %f %f\n",
                  vert->vertex.x, vert->vertex.y, vert->vertex.z,
                  vert->normal.x, vert->normal.y, vert->normal.z,
                  vert->uv.u, vert->uv.v);
         }

         //Write indices
         for (i=0; i<mNumIndices; i++) //For each index
         {
            fprintf(fp, "I %d\n", mIndexArray[i]);
         }

         //Close file
         fclose(fp);
      }
   }

   /***************/
   /* Free memory */
   /***************/
   if (mPolygonArray != NULL)
   {
      free(mPolygonArray);
      mPolygonArray = NULL;
   }

   /********************/
   /* End progress bar */
   /********************/
   ip->ProgressEnd();

   return 1;
}

//=============================================================================
BOOL SushiOut::CountAllByNodeEnum(INode* node) 
{
   mTotalNodeCount++;

   //Stop recursing if the user pressed Cancel
   if (ip->GetCancel())
      return FALSE;
   
   ObjectState os = node->EvalWorldState(0); 
   if (os.obj) //The os.obj is the actual object we will export
   {
      //We look at the super class ID to determine the type of the object.
      switch(os.obj->SuperClassID())
      {
         case GEOMOBJECT_CLASS_ID: 
            CountGeomObject(node); 
            break;
         case CAMERA_CLASS_ID:
            //CountCameraObject(node); 
            break;
         case LIGHT_CLASS_ID:
            //CountLightObject(node); 
            break;
         case SHAPE_CLASS_ID:
            //CountShapeObject(node); 
            break;
         case HELPER_CLASS_ID:
            //CountHelperObject(node); 
            break;
      }
   }

   //For each child of this node, we recurse into ourselves 
   //until no more children are found.
   for (int i=0; i<node->NumberOfChildren(); i++)
   {
      if (!CountAllByNodeEnum(node->GetChildNode(i)))
         return FALSE;
   }

   return TRUE;
}

//=============================================================================
// Count GeomObjects
//=============================================================================
void SushiOut::CountGeomObject(INode* node)
{
   ObjectState os = node->EvalWorldState(0);
   if ((!os.obj) || (os.obj->SuperClassID() != GEOMOBJECT_CLASS_ID))
      return;

   //Targets are actually geomobjects, but we will export them
   //from the camera and light objects, so we skip them here.
   if (os.obj->ClassID() == Class_ID(TARGET_CLASS_ID, 0))
      return;
   
   //Get triangle-based version of object for exporting
   BOOL needDel = FALSE;
   TriObject* tri = GetTriObjectFromNode(node, 0, needDel);
   if (!tri)
      return;
   
   /****************************************/
   /* Get mesh object and transform matrix */
   /****************************************/
   Mesh* mesh = &(tri->GetMesh());
   Matrix3 tm = node->GetObjTMAfterWSM(0);

   mNumPolygons += mesh->getNumFaces();

   /******************************/
   /* Delete temporary TriObject */
   /******************************/
   if (needDel)
      delete tri;
}

//=============================================================================
// Count Lights
//=============================================================================
void SushiOut::CountLightObject(INode* node)
{

}

//=============================================================================
// Count Helper Objects
//=============================================================================
void SushiOut::CountHelperObject(INode* node)
{

}

//=============================================================================
// This method is the main object exporter.
// It is called once of every node in the scene. The objects are
// exported as they are encoutered.
//
// Before recursing into the children of a node, we will export it.
// The benefit of this is that a nodes parent is always before the
// children in the resulting file. This is desired since a child's
// transformation matrix is optionally relative to the parent.
//=============================================================================
BOOL SushiOut::NodeEnum(INode* node) 
{
   mCurNode++;
   ip->ProgressUpdate((int)((float)mCurNode/mTotalNodeCount*100.0f)); 

   //Stop recursing if the user pressed Cancel
   if (ip->GetCancel())
      return FALSE;
   
   if (node->IsHidden() == TRUE)
      return TRUE;

   //The ObjectState is a 'thing' that flows down the pipeline containing
   //all information about the object. By calling EvalWorldState() we tell
   //max to eveluate the object at end of the pipeline.
   ObjectState os = node->EvalWorldState(0); 
   if (os.obj) //The os.obj is the actual object we will export
   {
      //We look at the super class ID to determine the type of the object.
      switch(os.obj->SuperClassID())
      {
         case GEOMOBJECT_CLASS_ID: 
            ExportGeomObject(node); 
            break;
         case CAMERA_CLASS_ID:
            //ExportCameraObject(node); 
            break;
         case LIGHT_CLASS_ID:
            //ExportLightObject(node); 
            break;
         case SHAPE_CLASS_ID:
            //ExportShapeObject(node); 
            break;
         case HELPER_CLASS_ID:
            //ExportHelperObject(node); 
            break;
      }
   }

   //For each child of this node, we recurse into ourselves 
   //until no more children are found.
   for (int i=0; i<node->NumberOfChildren(); i++)
   {
      if (!NodeEnum(node->GetChildNode(i)))
         return FALSE;
   }

   return TRUE;
}

//=============================================================================
// Global variables...currently does nothing
//=============================================================================
void SushiOut::ExportGlobalInfo()
{

}

//=============================================================================
// Export GeomObjects
//=============================================================================
void SushiOut::ExportGeomObject(INode* node)
{
   int i;
   float32 tmpf;

   ObjectState os = node->EvalWorldState(0);
   if ((!os.obj) || (os.obj->SuperClassID() != GEOMOBJECT_CLASS_ID))
      return;

   //Targets are actually geomobjects, but we will export them
   //from the camera and light objects, so we skip them here.
   if (os.obj->ClassID() == Class_ID(TARGET_CLASS_ID, 0))
      return;
   
   //Get triangle-based version of object for exporting
   BOOL needDel = FALSE;
   TriObject* tri = GetTriObjectFromNode(node, 0, needDel);
   if (!tri)
      return;
   
   /****************************************/
   /* Get mesh object and transform matrix */
   /****************************************/
   Mesh* mesh = &(tri->GetMesh());
   Matrix3 tm = node->GetObjTMAfterWSM(0);
   //node->GetName()
   //mesh->getNumVerts()
   //mesh->getNumFaces()   

   /***********************************/
   /* Compute Vertex and Face Normals */
   /***********************************/
   Tab<VNormal> vnorms;
   Tab<Point3> fnorms;
   if (1)
   {
      Point3 v0, v1, v2;
      int i;
      int j;

      vnorms.SetCount(mesh->getNumVerts());
      fnorms.SetCount(mesh->getNumFaces());

      //Initialize array
      for (i=0; i<mesh->getNumVerts(); i++)
      {
         vnorms[i] = VNormal();
      }

      //Compute face and vertex surface normals
      for (i=0; i<mesh->getNumFaces(); i++)
      {
         /********************/
         /* Ignore if hidden */
         /********************/
         if (mesh->faces[i].Hidden() == TRUE)
            continue;

         //Calculate the face normal
         v0 = (tm * mesh->verts[mesh->faces[i].v[0]]);
         v1 = (tm * mesh->verts[mesh->faces[i].v[1]]);
         v2 = (tm * mesh->verts[mesh->faces[i].v[2]]);
         //v0 = verts[face->v[0]];
         //v1 = verts[face->v[1]];
         //v2 = verts[face->v[2]];
         fnorms[i] = (v1-v0)^(v2-v1);
         
         //Normalize face normal
         fnorms[i] = Normalize(fnorms[i]);

         //Compute weighted vertex normals based on angle of vertex and accumulate it
         for (j=0; j<3; j++)
         {
            Point3 tmpPoint;

            tmpPoint = fnorms[i] * AtiMAXFindVertexAngle(mesh, i, j);
            vnorms[mesh->faces[i].v[j]].AddNormal(tmpPoint, mesh->faces[i].smGroup);
         }
      }

      //Normalize all vertex normals
      for (i=0; i<mesh->getNumVerts(); i++)
      {
         vnorms[i].Normalize();
      }
   }

   /*******************/
   /* Export polygons */
   /*******************/
   for (i=0; i<mesh->getNumFaces(); i++) //For each polygon in this object
   {
      /************************************/
      /* Pointer to current AtiPrePolygon */
      /************************************/
      NPatchPolygon *poly = &(mPolygonArray[mNumPolygons]);

      /********************/
      /* Ignore if hidden */
      /********************/
      if (mesh->faces[i].Hidden() == TRUE)
         continue;

      /**********************/
      /* Get vertex indices */
      /**********************/
      int v0 = mesh->faces[i].v[0];
      int v1 = mesh->faces[i].v[1];
      int v2 = mesh->faces[i].v[2];
      //mesh->faces[i].OrderVerts(v0, v1);
      //if (v0 != mesh->faces[i].v[0])
      //   MessageBox(NULL, "test", "test", MB_OK);

      /****************/
      /* Get Vertices */
      /****************/
      poly->vertex[0].x = (tm * mesh->verts[v0]).x;
      poly->vertex[0].y = (tm * mesh->verts[v0]).y;
      poly->vertex[0].z = (tm * mesh->verts[v0]).z;

      poly->vertex[1].x = (tm * mesh->verts[v1]).x;
      poly->vertex[1].y = (tm * mesh->verts[v1]).y;
      poly->vertex[1].z = (tm * mesh->verts[v1]).z;

      poly->vertex[2].x = (tm * mesh->verts[v2]).x;
      poly->vertex[2].y = (tm * mesh->verts[v2]).y;
      poly->vertex[2].z = (tm * mesh->verts[v2]).z;

      /**********************/
      /* Get vertex normals */
      /**********************/
      if (mesh->faces[i].smGroup != 0)
      {
         poly->normal[0].x = vnorms.Addr(v0)->GetNormal(mesh->faces[i].smGroup).x;
         poly->normal[0].y = vnorms.Addr(v0)->GetNormal(mesh->faces[i].smGroup).y;
         poly->normal[0].z = vnorms.Addr(v0)->GetNormal(mesh->faces[i].smGroup).z;
         tmpf = 1.0f / (float32)sqrt(poly->normal[0].x*poly->normal[0].x + poly->normal[0].y*poly->normal[0].y + poly->normal[0].z*poly->normal[0].z);
         poly->normal[0].x *= tmpf;
         poly->normal[0].y *= tmpf;
         poly->normal[0].z *= tmpf;

         poly->normal[1].x = vnorms.Addr(v1)->GetNormal(mesh->faces[i].smGroup).x;
         poly->normal[1].y = vnorms.Addr(v1)->GetNormal(mesh->faces[i].smGroup).y;
         poly->normal[1].z = vnorms.Addr(v1)->GetNormal(mesh->faces[i].smGroup).z;
         tmpf = 1.0f / (float32)sqrt(poly->normal[1].x*poly->normal[1].x + poly->normal[1].y*poly->normal[1].y + poly->normal[1].z*poly->normal[1].z);
         poly->normal[1].x *= tmpf;
         poly->normal[1].y *= tmpf;
         poly->normal[1].z *= tmpf;
      
         poly->normal[2].x = vnorms.Addr(v2)->GetNormal(mesh->faces[i].smGroup).x;
         poly->normal[2].y = vnorms.Addr(v2)->GetNormal(mesh->faces[i].smGroup).y;
         poly->normal[2].z = vnorms.Addr(v2)->GetNormal(mesh->faces[i].smGroup).z;
         tmpf = 1.0f / (float32)sqrt(poly->normal[2].x*poly->normal[2].x + poly->normal[2].y*poly->normal[2].y + poly->normal[2].z*poly->normal[2].z);
         poly->normal[2].x *= tmpf;
         poly->normal[2].y *= tmpf;
         poly->normal[2].z *= tmpf;
      }
      else
      {
         poly->normal[0].x = fnorms[i].x;
         poly->normal[0].y = fnorms[i].y;
         poly->normal[0].z = fnorms[i].z;

         poly->normal[1].x = fnorms[i].x;
         poly->normal[1].y = fnorms[i].y;
         poly->normal[1].z = fnorms[i].z;

         poly->normal[2].x = fnorms[i].x;
         poly->normal[2].y = fnorms[i].y;
         poly->normal[2].z = fnorms[i].z;
      }

      /****************/
      /* Get material */
      /****************/
      /*
      Mtl* nodeMtl = node->GetMtl();
      Mtl* faceMtl = NULL;
      if (nodeMtl != NULL)
         {
         if (nodeMtl->ClassID() == Class_ID(MULTI_CLASS_ID, 0))
            {
            faceMtl = nodeMtl->GetSubMtl(mesh->faces[i].getMatID() % nodeMtl->NumSubMtls());
            }
         else
            {
            faceMtl = nodeMtl;
            }
         }
      poly->materialID = SuAddMaterialToScene(faceMtl, &s);

      //Cancel export if problem reading material
      if (poly->materialID == -1)
      {
         ip->SetCancel(TRUE);
         break; //Stop exporting this mesh
      }
      */

      /*****************************************/
      /* Get texture coordinates if they exist */
      /*****************************************/
      UVVert *tv;
      TVFace *tf;
      tv = mesh->mapVerts(1);
      tf = mesh->mapFaces(1);
      if ((tv != NULL) && (tf != NULL))
      {
         poly->uv[0].u = tv[tf[i].t[0]].x;
         poly->uv[0].v = tv[tf[i].t[0]].y;
         poly->uv[1].u = tv[tf[i].t[1]].x;
         poly->uv[1].v = tv[tf[i].t[1]].y;
         poly->uv[2].u = tv[tf[i].t[2]].x;
         poly->uv[2].v = tv[tf[i].t[2]].y;
      }

      /*********************/
      /* Increment counter */
      /*********************/
      mNumPolygons++;
   }

   /******************************/
   /* Delete temporary TriObject */
   /******************************/
   if (needDel)
      delete tri;
}

//=============================================================================
// Shape output
//=============================================================================
void SushiOut::ExportShapeObject(INode* node)
{
   // No shape export
   // To export shapes they need to be capped and then they will be
   // exported as a tri-object
}

//=============================================================================
// Light output
//=============================================================================
void SushiOut::ExportLightObject(INode* node)
{

}

//=============================================================================
// Camera output
//=============================================================================
float FOVtoMM (float aperture, float fov)
{
   return (float)((0.5f*aperture)/tan(fov/2.0f));
}

//=============================================================================
void SushiOut::ExportCameraObject(INode* node)
{

}

//=============================================================================
// Helper object output
//=============================================================================
void SushiOut::ExportHelperObject(INode* node)
{
   //No Helper object export
}

//=============================================================================
// Node Header
//=============================================================================
void SushiOut::ExportNodeHeader(INode* node, TCHAR* type)
{
   //F the header.
}

//=============================================================================
// Return a pointer to a TriObject given an INode or return NULL
// if the node cannot be converted to a TriObject
//=============================================================================
TriObject* SushiOut::GetTriObjectFromNode(INode *node, TimeValue t, int &deleteIt)
{
   deleteIt = FALSE;

   Object *obj = node->EvalWorldState(t).obj;
   if (obj->CanConvertToType(Class_ID(TRIOBJ_CLASS_ID, 0)))
   { 
      TriObject *tri = (TriObject *) obj->ConvertToType(t, Class_ID(TRIOBJ_CLASS_ID, 0));

      //If a different pointer is given to tri, the memory needs to be freed
      //when done using it. ConvertToType() will create a new object if the
      //current object can't be treated as a TriObject.
      if (obj != tri)
         deleteIt = TRUE;

      return tri;
   }
   else
   {
      return NULL;
   }
}

//=============================================================================
SushiOut::SushiOut()
{

}

//=============================================================================
SushiOut::~SushiOut()
{

}

//=============================================================================
void SushiOut::AddVertexToArray (NPatchPolygon *poly, int32 idx)
{
   int32 i;
   int32 indexToAdd = -1;

   /***********************************/
   /* Do we already have this vertex? */
   /***********************************/
   for (i=0; i<mNumVertices; i++)
   {
      NPatchVertex *vert = &(mVertexArray[i]);

      if ((poly->vertex[idx].x == vert->vertex.x) &&
          (poly->vertex[idx].y == vert->vertex.y) &&
          (poly->vertex[idx].z == vert->vertex.z) &&
          (poly->normal[idx].x == vert->normal.x) &&
          (poly->normal[idx].y == vert->normal.y) &&
          (poly->normal[idx].z == vert->normal.z) &&
          (poly->uv[idx].u == vert->uv.u) &&
          (poly->uv[idx].v == vert->uv.v))
      {
         indexToAdd = i;
         break;
      }
   }

   /****************************************/
   /* Add vertex if we didn't find a match */
   /****************************************/
   if (indexToAdd == -1)
   {
      NPatchVertex *oldArray = NULL;
      NPatchVertex *vert = NULL;

      //Allocate memory for new vertex
      oldArray = mVertexArray;
      AtiMallocTypeArray (mVertexArray, mNumVertices+1, NPatchVertex);
      memcpy (mVertexArray, oldArray, mNumVertices * sizeof(NPatchVertex));
      AtiFreeTypeArray (oldArray);

      //Copy new vertex data to array
      vert = &(mVertexArray[mNumVertices]);
      vert->vertex.x = poly->vertex[idx].x;
      vert->vertex.y = poly->vertex[idx].y;
      vert->vertex.z = poly->vertex[idx].z;
      vert->normal.x = poly->normal[idx].x;
      vert->normal.y = poly->normal[idx].y;
      vert->normal.z = poly->normal[idx].z;
      vert->uv.u = poly->uv[idx].u;
      vert->uv.v = poly->uv[idx].v;

      //Increment counter
      indexToAdd = mNumVertices;
      mNumVertices++;
   }

   /*****************************/
   /* Add index to index buffer */
   /*****************************/
   uint32 *oldArray = mIndexArray;
   AtiMallocTypeArray (mIndexArray, mNumIndices+1, uint32);
   memcpy (mIndexArray, oldArray, mNumIndices * sizeof(uint32));
   AtiFreeTypeArray (oldArray);

   mIndexArray[mNumIndices] = indexToAdd;
   mNumIndices++;
}
